package at.ac.uibk.gitsearch.security.oauth2;

import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import java.io.IOException;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import javax.servlet.ServletException;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.oauth2.client.authentication.OAuth2AuthenticationToken;
import org.springframework.security.oauth2.core.oidc.user.DefaultOidcUser;
import org.springframework.security.oauth2.core.oidc.user.OidcUser;
import org.springframework.security.web.authentication.SavedRequestAwareAuthenticationSuccessHandler;

/**
 * allows for a redirect and adds a short-lived cookie that encodes a short lived JWT-Token
 * @author Michael Breu
 *
 */
public class SavedRequestAwareAuthenticationSuccessHandlerWithJWTSupport extends SavedRequestAwareAuthenticationSuccessHandler {

    protected TokenProvider tokenProvider;

    public SavedRequestAwareAuthenticationSuccessHandlerWithJWTSupport(TokenProvider tokenProvider) {
        super();
        this.tokenProvider = tokenProvider;
    }

    private static final int REQUEST_TOKEN_LIVETIME = 10; // seconds

    @Override
    public void onAuthenticationSuccess(HttpServletRequest request, HttpServletResponse response, Authentication authentication)
        throws ServletException, IOException {
        Authentication authenticationForToken = authentication;
        if (authentication instanceof OAuth2AuthenticationToken) {
            OAuth2AuthenticationToken oAuthA = (OAuth2AuthenticationToken) authentication;
            @SuppressWarnings("unchecked")
            List<String> gitLabGroups = (List<String>) ((OidcUser) ((OAuth2AuthenticationToken) authentication).getPrincipal()).getClaims()
                .get("groups");
            final DefaultOidcUser principal = (DefaultOidcUser) oAuthA.getPrincipal();
            List<GrantedAuthority> roles = new ArrayList<>();
            roles.addAll(authentication.getAuthorities());
            for (String gitLabGroup : gitLabGroups) {
                roles.add(new SimpleGrantedAuthority(gitLabGroup));
            }

            DefaultOidcUser emailUser = new DefaultOidcUser(roles, principal.getIdToken(), principal.getUserInfo(), "email");

            OAuth2AuthenticationToken oauth2Authentication = new OAuth2AuthenticationToken(
                emailUser,
                roles,
                oAuthA.getAuthorizedClientRegistrationId()
            );

            authenticationForToken = oauth2Authentication;
            authenticationForToken.setAuthenticated(authentication.isAuthenticated());
        }
        String token = tokenProvider.createToken(authenticationForToken, REQUEST_TOKEN_LIVETIME * 1000L, true/* preToken */); // 200 secs (for Debugging)

        Cookie tempTokenCookie = new Cookie("tempRequestToken", token);
        tempTokenCookie.setMaxAge(REQUEST_TOKEN_LIVETIME);
        tempTokenCookie.setPath("/");
        tempTokenCookie.setSecure(true);

        response.addCookie(tempTokenCookie);

        super.onAuthenticationSuccess(request, response, authentication);
    }

    @Override
    protected String determineTargetUrl(HttpServletRequest request, HttpServletResponse response, Authentication authentication) {
        return super.determineTargetUrl(request, response, authentication);
    }

    public static class SimplePrincipal implements Principal {

        protected String name;

        public SimplePrincipal(String name) {
            super();
            this.name = name;
        }

        @Override
        public String getName() {
            return name;
        }
    }

    public static class SimpleAuthentication implements Authentication {

        /**
         *
         */
        private static final long serialVersionUID = -791646857551363545L;

        private Principal principal;

        Collection<? extends GrantedAuthority> authorities;

        public SimpleAuthentication(Principal principal, Collection<? extends GrantedAuthority> authorities) {
            super();
            this.principal = principal;
            this.authorities = authorities;
        }

        @Override
        public String getName() {
            return principal.getName();
        }

        @Override
        public Collection<? extends GrantedAuthority> getAuthorities() {
            return authorities;
        }

        @Override
        public Object getCredentials() {
            return null;
        }

        @Override
        public Object getDetails() {
            return null;
        }

        @Override
        public Object getPrincipal() {
            return principal;
        }

        private boolean authenticated = false;

        @Override
        public boolean isAuthenticated() {
            return authenticated;
        }

        @Override
        public void setAuthenticated(boolean isAuthenticated) {
            authenticated = isAuthenticated;
        }
    }
}
