package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider.GitLabAccessInfo;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import org.apache.commons.lang3.StringUtils;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.codeability.sharing.plugins.api.ShoppingBasket;
import org.codeability.sharing.plugins.api.ShoppingBasket.UserInfo;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
public class ShoppingBasketService {

    @Autowired
    private PluginManagementService pluginManagementService;

    @Autowired
    private GitLabRepository gitLabRepository;

    public static class ShoppingBasketRedirectInfoDTO {

        private String redirectURL;

        /**
         * @return the redirectURL
         */
        public String getRedirectURL() {
            return redirectURL;
        }

        /**
         * @param redirectURL the redirectURL to set
         */
        public void setRedirectURL(String redirectURL) {
            this.redirectURL = redirectURL;
        }
    }

    public static class ShoppingBasketInfoDTO {

        private SearchResultDTO[] itemInfos;
        private String plugin;
        private String action;

        public ShoppingBasketInfoDTO(SearchResultDTO[] itemInfos) {
            super();
            this.itemInfos = itemInfos;
        }

        public ShoppingBasketInfoDTO() {
            // JSON
        }

        /**
         * @return the itemInfos
         */
        public SearchResultDTO[] getItemInfos() {
            return itemInfos;
        }

        /**
         * @param itemInfos the itemInfos to set
         */
        public void setItemInfos(SearchResultDTO[] itemInfos) {
            this.itemInfos = itemInfos;
        }

        /**
         * @return the plugin
         */
        public String getPlugin() {
            return plugin;
        }

        /**
         * @param plugin the plugin to set
         */
        public void setPlugin(String plugin) {
            this.plugin = plugin;
        }

        /**
         * @return the action
         */
        public String getAction() {
            return action;
        }

        /**
         * @param action the action to set
         */
        public void setAction(String action) {
            this.action = action;
        }

        public static class ItemInfoDTO {

            String exerciseID;
            String gitLabURL;

            public ItemInfoDTO(String exerciseID, String gitLabURl) {
                super();
                this.exerciseID = exerciseID;
                this.gitLabURL = gitLabURl;
            }

            public ItemInfoDTO() {
                // JSON
            }

            /**
             * @return the exerciseIDs
             */
            public String getExerciseID() {
                return exerciseID;
            }

            /**
             * @param exerciseIDs the exerciseIDs to set
             */
            public void setExerciseID(String exerciseID) {
                this.exerciseID = exerciseID;
            }

            /**
             * @return the gitLabURls
             */
            public String getGitLabURL() {
                return gitLabURL;
            }

            /**
             * @param gitLabURL the gitLabURls to set
             */
            public void setGitLabURL(String gitLabURL) {
                this.gitLabURL = gitLabURL;
            }
        }
    }

    /**
     * holds further info for shopping basket.
     * @author Michael Breu
     *
     */
    public static class ExtendedShoppingBasket {

        public ExtendedShoppingBasket(
            ShoppingBasketInfoDTO shoppingBasket,
            Optional<GitLabAccessInfo> gitLabAccessInfo,
            long tokenValidUntil
        ) {
            super();
            this.shoppingBasket = shoppingBasket;
            this.gitLabAccessInfo = gitLabAccessInfo;
            this.tokenValidUntil = tokenValidUntil;
        }

        public ExtendedShoppingBasket() {}

        private ShoppingBasketInfoDTO shoppingBasket;
        private Optional<GitLabAccessInfo> gitLabAccessInfo;

        /**
         * token valid until (msecs since 1.1.1970)
         */
        private long tokenValidUntil;

        /**
         * @return the shoppingBasket
         */
        public ShoppingBasketInfoDTO getShoppingBasket() {
            return shoppingBasket;
        }

        /**
         * @param shoppingBasket the shoppingBasket to set
         */
        public void setShoppingBasket(ShoppingBasketInfoDTO shoppingBasket) {
            this.shoppingBasket = shoppingBasket;
        }

        /**
         * @return the gitLabAccessInfo
         */
        public Optional<GitLabAccessInfo> getGitLabAccessInfo() {
            return gitLabAccessInfo;
        }

        /**
         * @param gitLabAccessInfo the gitLabAccessInfo to set
         */
        public void setGitLabAccessInfo(Optional<GitLabAccessInfo> gitLabAccessInfo) {
            this.gitLabAccessInfo = gitLabAccessInfo;
        }

        /**
         * @return the tokenValidUntil
         */
        public long getTokenValidUntil() {
            return tokenValidUntil;
        }
    }

    private final Logger log = LoggerFactory.getLogger(ShoppingBasketService.class);

    public ShoppingBasketService() {
        // JSON
    }

    public ShoppingBasket getBasket(String basketToken) {
        if (basketToken == null) {
            log.warn("Basket for null not found");
            return null;
        }
        @Nullable
        final ExtendedShoppingBasket basketInfo = basketCache.getIfPresent(basketToken);
        if (basketInfo == null) {
            log.warn("Basket {} not found", basketToken);
            return null;
        }

        List<org.codeability.sharing.plugins.api.search.SearchResultDTO> result = new ArrayList<>();
        for (SearchResultDTO hit : basketInfo.getShoppingBasket().getItemInfos()) {
            result.add(convertFrom(hit));
        }
        UserInfo userInfo = null;
        final Optional<GitLabAccessInfo> gitLabAccessInfo = basketInfo.getGitLabAccessInfo();
        if (gitLabAccessInfo.isPresent() && gitLabAccessInfo.get().getUser().isPresent()) {
            userInfo = new UserInfo(gitLabAccessInfo.get().getUser().get().getUsername());
        }
        final ShoppingBasket shoppingBasket = new ShoppingBasket(
            userInfo,
            result.toArray(new org.codeability.sharing.plugins.api.search.SearchResultDTO[] {}),
            basketInfo.tokenValidUntil
        );
        log.info("Basket {} is delivered", basketToken);
        return shoppingBasket;
    }

    static ObjectMapper objectMapper = new ObjectMapper();

    static {
        objectMapper.registerModule(new JavaTimeModule());
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    }

    private org.codeability.sharing.plugins.api.search.SearchResultDTO convertFrom(SearchResultDTO hit) {
        try {
            return objectMapper.readValue(
                objectMapper.writeValueAsString(hit),
                org.codeability.sharing.plugins.api.search.SearchResultDTO.class
            );
        } catch (JsonMappingException e) {
            log.warn("Cannot map hit", e);
            return null;
        } catch (JsonProcessingException e) {
            log.warn("Cannot convert hit", e);
            return null;
        }
    }

    public InputStream getRepositoryZip(String basketToken, int position) throws GitLabApiException, IOException {
        @Nullable
        final ExtendedShoppingBasket basketInfo = basketCache.getIfPresent(basketToken);
        ShoppingBasket basket = getBasket(basketToken);
        final GitLabApi gitLabApi = basketInfo == null
            ? gitLabRepository.getGitLabApi(Optional.empty())
            : gitLabRepository.getGitLabApi(basketInfo.getGitLabAccessInfo());
        return rePackageGitLabProjectZip(
            new ZipInputStream(
                gitLabApi.getRepositoryApi().getRepositoryArchive(basket.exerciseInfo[position].getProject().getProjectId(), "HEAD", "zip")
            ),
            "from project " + basket.exerciseInfo[position].getProject().getProjectId(),
            new String[] {},
            null
        );
    }

    /**
     * chops of the main folder, and brings every file one level up. Also deleting
     * all plain files in main folder.
     *
     * @param zipIn the zip to be repackaged
     * @param originalLocation the original location. For debug purposes only.
     * @param filterOut an array of file or foldernames that should be filtered away
     * @param exerciseFolderPath the path to the exercises
     * @return
     * @throws IOException
     */
    protected ZipInputStream rePackageGitLabProjectZip(
        ZipInputStream zipIn,
        String originalLocation,
        String[] filterOut,
        String exerciseFolderPath
    ) throws IOException {
        final PipedInputStream pis = new PipedInputStream(1024 * 256);
        final PipedOutputStream pos = new PipedOutputStream(pis);

        Runnable rePackage = () -> {
            try (ZipOutputStream zipOut = new ZipOutputStream(pos)) {
                ZipEntry zipInEntry = zipIn.getNextEntry();
                String prefix = "";
                if (zipInEntry.isDirectory()) { // main directory is prefix to all entries
                    prefix = zipInEntry.getName();
                }

                while (zipInEntry != null) {
                    String fileName = zipInEntry.getName();
                    if (fileName.startsWith(prefix)) {
                        String newName = fileName.substring(prefix.length());
                        String fileWithSubPath = getWithoutPathPrefix(newName, exerciseFolderPath);
                        if (fileWithSubPath != null) { // process this file
                            newName = fileWithSubPath;
                            if (!StringUtils.isEmpty(newName) && !filterOut(newName, filterOut)) {
                                ZipEntry zipOutEntry = new ZipEntry(newName);
                                if (zipInEntry.isDirectory()) {
                                    log.debug("ignoring directory {}", fileName);
                                    zipOut.putNextEntry(zipOutEntry);
                                } else {
                                    zipOut.putNextEntry(zipOutEntry);
                                    StreamUtils.copy(zipIn, zipOut);
                                    zipOut.closeEntry();
                                }
                            }
                        }
                    }
                    zipInEntry = zipIn.getNextEntry();
                }
                zipIn.close();
                zipOut.close();
            } catch (IOException e) {
                log.error("Cannot rezip file from {}", originalLocation);
            }
        };

        new Thread(rePackage).start();

        return new ZipInputStream(pis);
    }

    public String getExerciseFolderPath(String exercisesPath) {
        if (exercisesPath == null) return "";
        int slashPos = exercisesPath.lastIndexOf('/');
        if (slashPos < 0) return ""; // root metadata
        return exercisesPath.substring(0, slashPos);
    }

    /**
     * shortens the fileName by the pathPrefix. Returns null, if fileName does not start with path prefix.
     * @param fileName the file name
     * @param pathPrefix the path prefix
     * @return
     */
    private String getWithoutPathPrefix(String fileName, String pathPrefix) {
        if (StringUtils.isEmpty(pathPrefix)) return fileName;
        if (fileName.startsWith(pathPrefix) && (fileName.length() == pathPrefix.length() || fileName.charAt(pathPrefix.length()) == '/')) {
            return fileName.substring(pathPrefix.length() + 1);
        }
        return null;
    }

    private boolean filterOut(String fileName, String[] filterOut) {
        for (String filter : filterOut) {
            if (fileName.startsWith(filter)) {
                return fileName.length() == filter.length() || fileName.charAt(filter.length()) == '/';
            }
        }
        return false;
    }

    public static final Duration BASKET_EXPIRY_INTERVAL = Duration.ofMinutes(60);
    /**
     * this is a cache for shopping baskets. Entries expires after BASKET_EXPIRY_INTERVAL.
     * TODO: Store in Database, to allow multiple application instances.
     */
    protected LoadingCache<String, ExtendedShoppingBasket> basketCache = CacheBuilder
        .newBuilder()
        .maximumSize(10000)
        .expireAfterWrite(BASKET_EXPIRY_INTERVAL)
        .build(
            new CacheLoader<String, ExtendedShoppingBasket>() {
                public ExtendedShoppingBasket load(String key) {
                    return null;
                }
            }
        );

    public ShoppingBasketRedirectInfoDTO getRedirectInfo(ShoppingBasketInfoDTO basket, String baseURL) {
        UUID random = UUID.randomUUID();
        long tokenValidUntil = System.currentTimeMillis() + BASKET_EXPIRY_INTERVAL.toMillis();
        basketCache.put(random.toString(), new ExtendedShoppingBasket(basket, gitLabRepository.getGitLabAccessInfo(), tokenValidUntil));

        ShoppingBasketRedirectInfoDTO result = new ShoppingBasketRedirectInfoDTO();

        result.setRedirectURL(
            pluginManagementService.getRedirectURL(basket.getPlugin(), basket.getAction()) +
            "/" +
            random.toString() +
            "?returnURL=" +
            baseURL +
            "&apiBaseURL=" +
            baseURL +
            "/api/pluginIF/v0.1"
        );

        return result;
    }
}
