package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.service.ShoppingBasketService.ShoppingBasketInfoDTO;
import at.ac.uibk.gitsearch.service.ShoppingBasketService.ShoppingBasketRedirectInfoDTO;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import org.codeability.sharing.plugins.api.ShoppingBasket;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.elasticsearch.node.NodeValidationException;
import org.junit.Assert;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.security.test.context.support.WithMockUser;

@SpringBootTest(classes = GitsearchApp.class)
@WithMockUser(value = "test", authorities = "sharing")
public class ShoppingBasketServiceIT {

    private static final String TEST_ZIP_LOCATION = "./testData/junit-quality-tests-exercise-master.zip";

    @Autowired
    protected ShoppingBasketService shoppingBasketService;

    @Autowired
    private SearchService searchService;

    private static final Logger log = LoggerFactory.getLogger(ShoppingBasketServiceIT.class);

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the ElasticSearchTestConfiguration
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @Test
    public void testSimpleRepackage() throws IOException {
        final InputStream resourceStream = this.getClass().getResourceAsStream(TEST_ZIP_LOCATION);

        Assert.assertNotNull("cannot find resource at " + TEST_ZIP_LOCATION, resourceStream);
        final ZipInputStream testZip = new ZipInputStream(resourceStream);

        final InputStream testRepackagedZipIS = shoppingBasketService.rePackageGitLabProjectZip(
            testZip,
            "from " + TEST_ZIP_LOCATION,
            new String[] {},
            ""
        );

        final ZipInputStream testRepackagedZip = new ZipInputStream(testRepackagedZipIS);

        ZipEntry entry = testRepackagedZip.getNextEntry();

        while (entry != null) {
            log.info("found entry {} (directory: {})", entry.getName(), entry.isDirectory());
            Assert.assertTrue("original folder not chopped off?", !entry.getName().startsWith("junit-quality-tests-exercise-master"));
            entry = testRepackagedZip.getNextEntry();
        }

        testRepackagedZip.close();
    }

    @Test
    public void testFilteredRepackage() throws IOException {
        final String[] filterOut = new String[] { "solution", "tests" };

        // first check whether this is good test data
        final ZipInputStream originalZip = new ZipInputStream(this.getClass().getResourceAsStream(TEST_ZIP_LOCATION));

        boolean found = false;
        for (ZipEntry entry = originalZip.getNextEntry(); entry != null; entry = originalZip.getNextEntry()) {
            found = found || isPathPrefix(filterOut, "junit-quality-tests-exercise-master/", entry);
            if (found) break;
        }
        Assert.assertTrue("test zip should contain entries to filter away?", found);

        final InputStream resourceStream = this.getClass().getResourceAsStream(TEST_ZIP_LOCATION);

        Assert.assertNotNull("cannot find resource at " + TEST_ZIP_LOCATION, resourceStream);
        final ZipInputStream testZip = new ZipInputStream(resourceStream);

        final InputStream testRepackagedZipIS = shoppingBasketService.rePackageGitLabProjectZip(
            testZip,
            "from " + TEST_ZIP_LOCATION,
            filterOut,
            ""
        );

        final ZipInputStream testRepackagedZip = new ZipInputStream(testRepackagedZipIS);

        for (ZipEntry entry = testRepackagedZip.getNextEntry(); entry != null; entry = testRepackagedZip.getNextEntry()) {
            boolean foundPrefix = isPathPrefix(filterOut, "", entry);
            Assert.assertFalse("original folder starts with excepted directory?", foundPrefix);
        }

        testRepackagedZip.close();
    }

    /**
     * tests for the path prefix in filterOut
     * @param filterOut the list of path prefixes to filter out
     * @param basePrefix a basePrefix to ignore
     * @param entry the zip entry
     * @return true if one of the path prefixes matches
     */
    private boolean isPathPrefix(final String[] filterOut, String basePrefix, ZipEntry entry) {
        boolean found = false;
        String fileName = entry.getName();
        if (fileName.startsWith(basePrefix)) fileName = fileName.substring(basePrefix.length());
        for (String filter : filterOut) {
            found = found || fileName.startsWith(filter); // this is slighlty lazy, because we should check for real path
            if (found) break;
        }
        return found;
    }

    /**
     * tests for subpackages
     * @throws IOException
     */
    @Test
    public void testSubRepackage() throws IOException {
        final InputStream resourceStream = this.getClass().getResourceAsStream(TEST_ZIP_LOCATION);

        Assert.assertNotNull("cannot find resource at " + TEST_ZIP_LOCATION, resourceStream);
        final ZipInputStream testZip = new ZipInputStream(resourceStream);

        final InputStream testRepackagedZipIS = shoppingBasketService.rePackageGitLabProjectZip(
            testZip,
            "from " + TEST_ZIP_LOCATION,
            new String[] {},
            "exercise"
        );

        final ZipInputStream testRepackagedZip = new ZipInputStream(testRepackagedZipIS);

        for (ZipEntry entry = testRepackagedZip.getNextEntry(); entry != null; entry = testRepackagedZip.getNextEntry()) {
            boolean foundPrefix = isPathPrefix(new String[] { "solution", "tests", "test" }, "", entry);
            Assert.assertFalse("original folder starts with excepted directory?", foundPrefix);
        }

        testRepackagedZip.close();
    }

    @Test
    public void testGetBasket() throws IOException {
        // just load everything from index :-)
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
        Assert.assertTrue("At least one hit?", searchResultPage.getSearchResult().size() >= 1);

        ShoppingBasketInfoDTO sbi = new ShoppingBasketInfoDTO(searchResultPage.getSearchResult().toArray(new SearchResultDTO[] {}));

        ShoppingBasketRedirectInfoDTO redInfo = shoppingBasketService.getRedirectInfo(sbi, "http://unused/unused");

        Pattern tokenParser = Pattern.compile("/([^/?]*)\\?");
        final Matcher matcher = tokenParser.matcher(redInfo.getRedirectURL());
        Assert.assertTrue(matcher.find());
        String token = matcher.group(1);
        final ShoppingBasket basket = shoppingBasketService.getBasket(token);

        int count = 0;
        for (SearchResultDTO hitResult : searchResultPage.getSearchResult()) {
            // we assume same sequence
            Assert.assertEquals(basket.exerciseInfo[count].getMetadata().getTitle(), hitResult.getMetadata().getTitle());
            count++;
        }
    }
}
