package at.ac.uibk.gitsearch.web.rest;

import static org.hamcrest.Matchers.hasProperty;
import static org.hamcrest.Matchers.is;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.web.rest.ExerciseResourceIT.StreamResultMatcher;
import java.io.IOException;
import java.util.concurrent.TimeUnit;
import org.elasticsearch.node.NodeValidationException;
import org.hamcrest.MatcherAssert;
import org.hamcrest.Matchers;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;

/**
 * Integration tests for the {@link UserResource} REST controller.
 */
@AutoConfigureMockMvc
@WithMockUser(authorities = { AuthoritiesConstants.USER, "sharing" })
@SpringBootTest(classes = GitsearchApp.class)
public class SearchResourceIT {

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the ElasticSearchTestConfiguration
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @Autowired
    private MockMvc restSearchMockMvc;

    @Test
    public void testKeywordsAutoComplete() throws Exception {
        restSearchMockMvc
            .perform(
                get("/api/search/keywordsAutoComplete")
                    .param("keyWordPrefix", "Jav")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andExpect(content().contentType(MediaType.APPLICATION_JSON))
            .andExpect(status().isOk())
            .andExpect(
                TestUtil.testResult(
                    AutoCompleteEntry[].class,
                    r -> {
                        MatcherAssert.assertThat(r, Matchers.arrayContaining(hasProperty("target", is("Java"))));
                    }
                )
            );
    }

    @Test
    public void testCreatorContributorAutoComplete() throws Exception {
        restSearchMockMvc
            .perform(
                get("/api/search/contributorCreatorAutoComplete")
                    .param("contributorPrefix", "Pod")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andExpect(content().contentType(MediaType.APPLICATION_JSON))
            .andExpect(status().isOk())
            .andExpect(
                TestUtil.testResult(
                    AutoCompleteEntry[].class,
                    r -> {
                        MatcherAssert.assertThat(r, Matchers.hasItemInArray(hasProperty("target", is("Stefan Podlipnig"))));
                    }
                )
            );
    }

    @Test
    public void testProgrammingLanguageAutocompletion() throws Exception {
        restSearchMockMvc
            .perform(
                get("/api/search/programmingLanguageAutoComplete")
                    .param("programmingLanguagePrefix", "Ja")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andExpect(content().contentType(MediaType.APPLICATION_JSON))
            .andExpect(status().isOk())
            .andExpect(
                TestUtil.testResult(
                    AutoCompleteEntry[].class,
                    r -> {
                        MatcherAssert.assertThat(r, Matchers.hasItemInArray(hasProperty("target", is("JAVA"))));
                    }
                )
            );
    }

    @Test
    public void getREADME() throws Exception {
        //    	String exerciseId = "272"; // warning: this depends on the sequence in the current search index :-(
        //    	restSearchMockMvc.perform(post("/api/exerciseFiles/"+exerciseId)
        String exerciseId = "318"; // warning: this depends on the sequence in the current search index :-(
        restSearchMockMvc
            .perform(
                post("/api/programming-exercises/" + exerciseId + "/export-programming-exercise")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            //    		.andDo(print())
            .andExpect(status().is2xxSuccessful())
            .andExpect(StreamResultMatcher.expectOctetStream())
            .andExpect(StreamResultMatcher.expectMinimumSize(500));
    }
}
